// ---------------------------------------------------------------------------

// STK500 programmer: commands parser module

// ---------------------------------------------------------------------------

#include "main.h"
#include "cmd.h"
#include "command.h"
#include "terminal.h"
#include "uart.h"
#include "adc.h"
#include "spi.h"

#include <string.h>

// ----------------------------- Constants: ----------------------------------

// device name to replay for CMD_SIGN_ON command
#define DEV_STK "STK500_2" 
#define DEV_ISP "AVRISP_2"

#define CONFIG_PARAM_BUILD_NUMBER_LOW   0
#define CONFIG_PARAM_BUILD_NUMBER_HIGH  1
// be careful with changing HW versions.
// AVR Studio does not like all numbers.
#define CONFIG_PARAM_HW_VER             2
#define CONFIG_PARAM_VADJUST           25
#define CONFIG_PARAM_OSC_PSCALE         1
#define CONFIG_PARAM_OSC_CMATCH         1

#define MSG_TIMEOUT  250 // message timeout, ms

// ----------------------------- Variables: ----------------------------------

static char larger_than_64k = 0; // handling off addressed larger than 64k words
static char new_address = 0;
static char extended_address = 0;
static unsigned long address = 0;
static unsigned int saddress = 0;
static char msg_buf[295];
static char param_controller_init = 0;
static char detected_vtg_from_reset_pin = 0; // reset pin voltage, x0.1 V
static char prg_state = 0;    // 0 = idle, 1 = programming is ongoing
static char MsgParseState;    // state of message parser
static bool TerminalMode = 0; // 0 = cmd, 1 = term. Windows, 2 = term. Linux
static char OscPscale;
static char OscCmatch;

// ------------------------ Function prototypes: -----------------------------

void TxAnswer(char seqnum, unsigned int len);
void ProgCmd(char seqnum);
void Cmd_SetOscOut(void);

// --------------------------- Module Init: ----------------------------------

void Cmd_Init(void)
{
  Terminal_Init();
  Uart_Init();
  Spi_Init();
  MsgParseState = MSG_IDLE;
  Cmd_ResetOk();
  OscPscale = CONFIG_PARAM_OSC_PSCALE;
  OscCmatch = CONFIG_PARAM_OSC_CMATCH;
  Cmd_SetOscOut();
}

// -------------------------- Parse message: ---------------------------------

void Cmd_Parse(bool t)
{
  static int msglen = 0;
  static char seqnum = 0;
  static int i = 0;
  static char cksum = 0;
  static int MsgTimer = 0; // message timeout timer
  char ch;
  
  // check for incoming byte
  if(!Uart_Ready())
  {
    // no byte and message not complete
    if(MsgParseState != MSG_IDLE)
    {
      if(t)
      {
        // check timeout
        if(MsgTimer) MsgTimer--;
        else
        {
          // reset message
          Spi_Disable();
          MsgParseState = MSG_IDLE;
          msglen = 0;
          seqnum = 0;
          i = 0;
          cksum = 0;
        }
      }
    }
  }
  else
  {
    MsgTimer = MSG_TIMEOUT; // timeout reload
    ch = Uart_GetChar();
    if(MsgParseState == MSG_IDLE)
    {
      // try to enter in special terminal mode:
      TerminalMode = Terminal_Exe(ch);
      if(TerminalMode) return; // skip parsing in terminal mode
    }
    // parse message according to appl. note AVR068 table 3-1:
    if(MsgParseState == MSG_IDLE && ch == MESSAGE_START)
    {
      MsgParseState = MSG_WAIT_SEQNUM;
      cksum = ch ^ 0;
      return;
    }
    if(MsgParseState == MSG_WAIT_SEQNUM)
    {
      seqnum = ch;
      cksum ^= ch;
      MsgParseState = MSG_WAIT_SIZE1;
      return;
    }
    if(MsgParseState == MSG_WAIT_SIZE1)
    {
      cksum ^= ch;
      msglen = ch << 8;
      MsgParseState = MSG_WAIT_SIZE2;
      return;
    }
    if(MsgParseState == MSG_WAIT_SIZE2)
    {
      cksum ^= ch;
      msglen |= (ch & 0xff);
      MsgParseState = MSG_WAIT_TOKEN;
      return;
    }
    if(MsgParseState == MSG_WAIT_TOKEN)
    {
      cksum ^= ch;
      if(ch == TOKEN)
      {
        MsgParseState = MSG_WAIT_MSG;
        i = 0;
      }
      else
      {
        MsgParseState = MSG_IDLE;
      }
      return;
    }
    if(MsgParseState == MSG_WAIT_MSG && i < msglen && i < 280)
    {
      cksum ^= ch;
      msg_buf[i] = ch;
      i++;
      Main_Wdt_Reset(1);
      if(i == msglen)
        MsgParseState = MSG_WAIT_CKSUM;
      return;
    }
    if(MsgParseState == MSG_WAIT_CKSUM)
    {
      if(ch == cksum && msglen > 0)
      {
        // message correct, process it
        Main_Wdt_Reset(1);
        ProgCmd(seqnum);
      }
      else
      {
        msg_buf[0] = ANSWER_CKSUM_ERROR;
        msg_buf[1] = STATUS_CKSUM_ERROR;
        TxAnswer(seqnum, 2);
      }
      // no return here, set state = MSG_IDLE
    }
    MsgParseState = MSG_IDLE;
    msglen = 0;
    seqnum = 0;
    cksum = 0;
    i = 0;
  }
}

// ---------- Transmit an answer back to the programmer software -------------

// message is in msg_buf,
// seqnum is the seqnum of the last message from the programmer software,
// len = 1..275 according to avr068

void TxAnswer(char seqnum, unsigned int len)
{
  unsigned char cksum;
  unsigned char ch;
  int i;
  if(len > 285 || len < 1) // software error
  {
    len = 2;
    // msg_buf[0]: not changed
    msg_buf[1] = STATUS_CMD_FAILED;
  }
  Uart_SendChar(MESSAGE_START); // 0x1B
  cksum = MESSAGE_START ^ 0;
  Uart_SendChar(seqnum);
  cksum ^= seqnum;
  ch = (len >> 8) & 0xFF;
  Uart_SendChar(ch);
  cksum ^= ch;
  ch = len & 0xFF;
  Uart_SendChar(ch);
  cksum ^= ch;
  Uart_SendChar(TOKEN); // 0x0E
  cksum ^= TOKEN;
  Main_Wdt_Reset(1);
  for(i = 0; i < len; i++)
  {
    Uart_SendChar(msg_buf[i]);
    cksum ^= msg_buf[i];
  }
  Uart_SendChar(cksum);
}

// ---------------------------- Check RESET pin ------------------------------

bool Cmd_ResetOk(void)
{
  static bool rst_ok = 0;

  if(!prg_state)
  {
    if(Adc_Get() < RES_PIN_LEVEL) rst_ok = 0;
    else if(!rst_ok)
    {
      if(Terminal_GetResetCheck() == 0) rst_ok = 1;
      else
      {
        DIR_RESET_OUT; // reset pin is output
        PORT_RESET_0;  // reset = low
        Delay_ms(10);
        rst_ok = Adc_Get() < RES_PIN_LEVEL;
        DIR_RESET_IN; // reset as input, high impedance
      }
    }
  }
  return(rst_ok);
}

// ------------------------- Set Osc. Out frequency: -------------------------

void Cmd_SetOscOut(void)
{
  char p = OscPscale;
  char m = 1;
  switch(OscPscale)
  {
  case 3: p = 2; m = 4; break; // CK / 32 = CK / 8 / 4
  case 4: p = 3; break;        // CK / 64
  case 5: p = 3; m = 2; break; // CK / 128 = CK / 64 / 2
  case 6: p = 4; break;        // CK / 256
  case 7: p = 5; break;        // CK / 1024
  }
  TCCR1B = (1 << WGM12) | (p & 7); // CTC, prescaler
  if(p)
    TCCR1A = (1 << COM1A0); // toggle OC1A on compare
      else TCCR1A = 0;      // target clock off
  int c = ((int)OscCmatch + 1) * (long)(F_CLK * 1E6) / (long)(F_CLK_STK * 1E6);
  if(c < 1) c = 1;
  OCR1A = c * m - 1;
}

// -------------------------- Returns parser state: --------------------------

char Cmd_ParseState(void)
{
  return(MsgParseState);
}

// ------------------------ Check for terminal mode: -------------------------

bool Cmd_TerminalMode(void)
{
  return(TerminalMode);
}

// ---------------------------------------------------------------------------
// ----------------------- Commands implementation ---------------------------
// ---------------------------------------------------------------------------

void ProgCmd(char seqnum)
{
  unsigned char tmp, tmp2, addressing_is_word, ci, cj, cstatus;
  unsigned int answerlen;
  unsigned int poll_address = 0;
  unsigned int i, nbytes;
  // distingush addressing CMD_READ_EEPROM_ISP (8bit) and CMD_READ_FLASH_ISP (16bit)
  addressing_is_word = 1; // 16 bit is default

  switch(msg_buf[0])
  {
    
// ---------------------------- CMD_SIGN_ON ----------------------------------
  
  case CMD_SIGN_ON:
    // prepare answer:
    msg_buf[0] = CMD_SIGN_ON;   // 0x01
    msg_buf[1] = STATUS_CMD_OK; // 0x00
    msg_buf[2] = 8;             // len
    strcpy((char *)&(msg_buf[3]),DEV_STK);
    answerlen = 11;
    break;

// ------------------------- CMD_SET_PARAMETER -------------------------------

  case CMD_SET_PARAMETER:
    // not implemented:
    // PARAM_VTARGET
    // PARAM_VADJUST
    // PARAM_RESET_POLARITY
    switch(msg_buf[1])
    {
    case PARAM_OSC_PSCALE:
      OscPscale = msg_buf[2];
      Cmd_SetOscOut();
      break;
    case PARAM_OSC_CMATCH:
      OscCmatch = msg_buf[2];
      Cmd_SetOscOut();
      break;
    case PARAM_SCK_DURATION:
      Spi_SetSckDuration(msg_buf[2]);
      break;
    case PARAM_CONTROLLER_INIT:
      param_controller_init = msg_buf[2];
      break;
    }
    answerlen = 2;
    //msg_buf[0] = CMD_SET_PARAMETER;
    msg_buf[1] = STATUS_CMD_OK;
    break;

// ------------------------- CMD_GET_PARAMETER -------------------------------

  case CMD_GET_PARAMETER:
      tmp=0xff;
      switch(msg_buf[1])
      {
      case PARAM_BUILD_NUMBER_LOW:
        tmp = CONFIG_PARAM_BUILD_NUMBER_LOW;
        break;
      case PARAM_BUILD_NUMBER_HIGH:
        tmp = CONFIG_PARAM_BUILD_NUMBER_HIGH;
        break;
      case PARAM_HW_VER:
       tmp = CONFIG_PARAM_HW_VER;
        break;
      case PARAM_SW_MAJOR:
        tmp = Terminal_GetSWMajor();
        break;
      case PARAM_SW_MINOR:
        tmp = Terminal_GetSWMinor();
        break;
      case PARAM_VTARGET:
        if(detected_vtg_from_reset_pin)
          tmp = detected_vtg_from_reset_pin; // unit: voltage*10: 50 means 5V
            else tmp = Adc_Get(); // measure reset pin
        break;
      case PARAM_VADJUST:
        tmp = CONFIG_PARAM_VADJUST;
        break;
      case PARAM_SCK_DURATION:
        tmp = Spi_GetSckDuration();
        break;
      case PARAM_RESET_POLARITY:   // is actually write only, list anyhow
        // 1 = avr (reset active = low), 0 = at89 (not supported by avrusb500)
        tmp = 1;
        break;
      case PARAM_CONTROLLER_INIT:
        tmp = param_controller_init;
        break;
      case PARAM_OSC_PSCALE:
        tmp = OscPscale;
        break;
      case PARAM_OSC_CMATCH:
        tmp = OscCmatch;
        break;
      case PARAM_TOPCARD_DETECT: // stk500 only
        tmp = 0x88;
        break;
      case PARAM_DATA: // stk500 only
        tmp = 0; 
        break;
      default: 
        break;
      }
    if(tmp == 0xff)
    {
      // command not understood
      answerlen = 2;
      //msg_buf[0] = CMD_GET_PARAMETER;
      msg_buf[1] = STATUS_CMD_UNKNOWN;
    }
    else
    {
      answerlen = 3;
      //msg_buf[0] = CMD_GET_PARAMETER;
      msg_buf[1] = STATUS_CMD_OK;
      msg_buf[2] = tmp;
    }
    break;

// -------------------------- CMD_LOAD_ADDRESS -------------------------------

  case CMD_LOAD_ADDRESS:
    address =  ((unsigned long)msg_buf[1]) << 24;
    address |= ((unsigned long)msg_buf[2]) << 16;
    address |= ((unsigned long)msg_buf[3]) << 8;
    address |= ((unsigned long)msg_buf[4]);
    // atmega2561/atmega2560
    // If bit 31 is set, this indicates that the following read/write operation
    // will be performed on a memory that is larger than 64KBytes. This is an
    // indication to STK500 that a load extended address must be executed. See
    // datasheet for devices with memories larger than 64KBytes.
    if(msg_buf[1] >= 0x80)
      larger_than_64k = 1;
        else larger_than_64k = 0;
    extended_address = msg_buf[2];
    new_address = 1;
    answerlen = 2;
    //msg_buf[0] = CMD_LOAD_ADDRESS;
    msg_buf[1] = STATUS_CMD_OK;
    break;

// ------------------------- CMD_FIRMWARE_UPGRADE ----------------------------

  case CMD_FIRMWARE_UPGRADE:
    // firmare upgrade is not supported this way
    answerlen = 2;
    //msg_buf[0] = CMD_FIRMWARE_UPGRADE;
    msg_buf[1] = STATUS_CMD_FAILED;
    break;

// ------------------------ CMD_ENTER_PROGMODE_ISP ---------------------------

  case CMD_ENTER_PROGMODE_ISP: // 0x10
    // The syntax of this command is as follows:
    // 0: Command ID 1 byte, CMD_ENTER_ PROGMODE_ISP
    // 1: timeout 1 byte, Command time-out (in ms)
    // 2: stabDelay 1 byte, Delay (in ms) used for pin stabilization
    // 3: cmdexeDelay 1 byte, Delay (in ms) in connection with the EnterProgMode command execution
    // 4: synchLoops 1 byte, Number of synchronization loops
    // 5: byteDelay 1 byte, Delay (in ms) between each byte in the EnterProgMode command.
    // 6: pollValue 1 byte, Poll value: 0x53 for AVR, 0x69 for AT89xx
    // 7: pollIndex 1 byte, Start address, received byte: 0 = no polling, 3 = AVR, 4 = AT89xx
    // cmd1 1 byte
    // cmd2 1 byte
    // cmd3 1 byte
    // cmd4 1 byte
    prg_state = 1;
    Spi_Disable();
    Delay_ms(15); 
    detected_vtg_from_reset_pin = Adc_Get(); // measure reset pin
    Spi_Enable();
    Delay_ms(msg_buf[2]); // stabDelay
    answerlen = 2;
    //msg_buf[0] = CMD_ENTER_PROGMODE_ISP;
    // set default to failed:
    msg_buf[1] = STATUS_CMD_FAILED;
    // connect to the target:
    i = 0;
    if(msg_buf[4] > 48) msg_buf[4] = 48; // limit the loops
    if(msg_buf[5] < 1)  msg_buf[5] = 1;  // minimum byteDelay
    while(i < msg_buf[4]) // synchLoops
    {
      Main_Wdt_Reset(1);
      Delay_ms(msg_buf[3]); // cmdexeDelay
      i++;
      Spi_TxRx(msg_buf[8]);         // cmd1
      Delay_ms(msg_buf[5]);         // byte delay
      Spi_TxRx(msg_buf[9]);         // cmd2
      Delay_ms(msg_buf[5]);         // byte delay
      tmp = Spi_TxRx(msg_buf[10]);  // cmd3
      Delay_ms(msg_buf[5]);         // byte delay
      tmp2 = Spi_TxRx(msg_buf[11]); // cmd4
      if(msg_buf[7] == 3 && tmp == msg_buf[6]) //7 = pollIndex, 6 = pollValue
        msg_buf[1] = STATUS_CMD_OK;
      if(msg_buf[7]!=3 && tmp2==msg_buf[6])//7 = pollIndex, 6 = pollValue
        msg_buf[1] = STATUS_CMD_OK;
      if(msg_buf[7] == 0) //pollIndex
        msg_buf[1] = STATUS_CMD_OK;
      if(msg_buf[1] == STATUS_CMD_OK)
      {
        i = msg_buf[4];// end loop
      }
      else
      {
        // new try, see e.g chapeter "Serial download" in data sheet.
        // in atmega8 it says: give positive reset pulse and try again
        // in at90s2313 it says: give positive sck pulse and try again
        // One of those datasheets must be wrong!?
        // Guido thinks that spi_sck_pulse is correct
        // and will also work for atmega8 given that
        // the number of synchLoops is at least 32.
        //Spi_RESET_Pulse();
        Spi_SCK_Pulse();
        Delay_ms(20); 
      }
      if(msg_buf[1] == STATUS_CMD_FAILED)
        prg_state = 0;
    }
    break;
          
// ------------------------ CMD_LEAVE_PROGMODE_ISP ---------------------------

  case CMD_LEAVE_PROGMODE_ISP:
    prg_state = 0;
    Spi_Disable();
    detected_vtg_from_reset_pin = 0;
    answerlen = 2;
    //msg_buf[0] = CMD_LEAVE_PROGMODE_ISP;
    msg_buf[1] = STATUS_CMD_OK;
    break;

// -------------------------- CMD_CHIP_ERASE_ISP -----------------------------

  case CMD_CHIP_ERASE_ISP:
    Spi_TxRx(msg_buf[3]);
    Spi_TxRx(msg_buf[4]);
    Spi_TxRx(msg_buf[5]);
    Spi_TxRx(msg_buf[6]);
    if(msg_buf[2] == 0)
    {
      // pollMethod use delay
      Delay_ms(msg_buf[1]); // eraseDelay
    }
    else
    {
      // pollMethod RDY/BSY cmd
      ci = 150; // timeout
      do
      {
        ci--;
        Spi_TxRx(0xF0);
        Spi_TxRx(0);
        Spi_TxRx(0);
      }
      while((Spi_TxRx(0) & 1) && ci);
    }
    answerlen = 2;
    //msg_buf[0] = CMD_CHIP_ERASE_ISP;
    msg_buf[1] = STATUS_CMD_OK;
    break;

// ------------------------ CMD_PROGRAM_EEPROM_ISP ---------------------------

  case CMD_PROGRAM_EEPROM_ISP:
    addressing_is_word = 0;  // address each byte
    case CMD_PROGRAM_FLASH_ISP:
    // msg_buf[0] CMD_PROGRAM_FLASH_ISP = 0x13
    // msg_buf[1] NumBytes H
    // msg_buf[2] NumBytes L
    // msg_buf[3] mode
    // msg_buf[4] delay
    // msg_buf[5] cmd1 (Load Page, Write Program Memory)
    // msg_buf[6] cmd2 (Write Program Memory Page)
    // msg_buf[7] cmd3 (Read Program Memory)
    // msg_buf[8] poll1 (value to poll)
    // msg_buf[9] poll2
    // msg_buf[n+10] Data
    poll_address = 0;
    ci = 150;
    
    if(msg_buf[4] < 4)  msg_buf[4] = 4;  // set a minimum timed delay
    if(msg_buf[4] > 32) msg_buf[4] = 32; // set a max delay
		saddress = (address & 0xFFFF);       // previous address, start address 
		nbytes = (unsigned int)((msg_buf[1] << 8) | msg_buf[2]);
    if(nbytes > 280)
    {
      // corrupted message
      answerlen = 2;
      msg_buf[1] = STATUS_CMD_FAILED;
      break;
    }
    // store the original mode:
    tmp2=msg_buf[3];
    // result code
    cstatus=STATUS_CMD_OK;
    // msg_buf[3] test Word/Page Mode bit:
    if((msg_buf[3] & 1) == 0)
    {
      // word mode
      for(i = 0; i < nbytes; i++)
      {        
        // The Low/High byte selection bit is
        // bit number 3. Set high byte for uneven bytes
        if(addressing_is_word && i & 1)               
          Spi_TxRx(msg_buf[5] | (1 << 3)); // high byte
            else Spi_TxRx(msg_buf[5]);     // low byte 
        Spi_TxRx(HI(address));
        Spi_TxRx(LO(address));
        Spi_TxRx(msg_buf[i + 10]);
        // if the data byte is not same as poll value
        // in that case we can do polling:
        if(msg_buf[8] != msg_buf[i + 10])
        {
          poll_address = address & 0xFFFF;
          // restore the possibly modifed mode:
          msg_buf[3] = tmp2;
        }
        else
        {
          //switch the mode to timed delay (waiting), for this word
          msg_buf[3] = 0x02;
        }
        Main_Wdt_Reset(1);
        if(!addressing_is_word) Delay_ms(2); // eeprom writing, eeprom needs more time
        //check the different polling mode methods
        if(msg_buf[3] & 0x04)
        {
          //data value polling
          tmp = msg_buf[8];
          ci = 150; // timeout
          while(tmp == msg_buf[8] && ci)
          {
            // The Low/High byte selection bit is
            // bit number 3. Set high byte for uneven bytes
            // Read data:
            if(addressing_is_word && i & 1)
              Spi_TxRx(msg_buf[7] | (1 << 3));
                else Spi_TxRx(msg_buf[7]);
            Spi_TxRx(HI(poll_address));
            Spi_TxRx(LO(poll_address));
            tmp = Spi_TxRx(0x00);
            ci--;
          }
        }
        else if(msg_buf[3] & 0x08)
        {
          //RDY/BSY polling
          ci = 150; // timeout
          do
          {
            ci--;
            Spi_TxRx(0xF0);
            Spi_TxRx(0);
            Spi_TxRx(0);
          }
          while((Spi_TxRx(0) & 1) && ci);
        }
        else
        {
            //timed delay (waiting)
            Delay_ms(msg_buf[4]);
        }
        if(addressing_is_word)
        {
          //increment word address only when we have an uneven byte
          if(i & 1) address++;
        }
        else
        {
          //increment address
          address++;
        }
				if(ci == 0)
        {
					 cstatus=STATUS_CMD_TOUT;
				}
      }                        
    }
    else
    {
      //page mode, all modern chips, atmega etc...
      i=0;
      while(i < nbytes)
      {
        Main_Wdt_Reset(1);
        // In commands PROGRAM_FLASH and READ_FLASH "Load Extended Address" 
        // command is executed before every operation if we are programming 
        // processor with Flash memory bigger than 64k words and 64k words boundary 
        // is just crossed or new address was just loaded.
        if(larger_than_64k && ((address & 0xFFFF) == 0 || new_address))
        {
          // load extended addr byte 0x4d
          Spi_TxRx(0x4d);
          Spi_TxRx(0x00);
          Spi_TxRx(extended_address);
          Spi_TxRx(0x00);
          new_address = 0;
        }
        // The Low/High byte selection bit is
        // bit number 3. Set high byte for uneven bytes
        if(addressing_is_word && i & 1)
          Spi_TxRx(msg_buf[5] | (1 << 3));
            else Spi_TxRx(msg_buf[5]);
        Spi_TxRx(HI(address));
        Spi_TxRx(LO(address));
        Spi_TxRx(msg_buf[i + 10]);
        // the data byte is not same as poll value
        // in that case we can do polling:
        if(msg_buf[8] != msg_buf[i + 10])
          poll_address = address & 0xFFFF;
            //switch the mode to timed delay (waiting)
            //we must preserve bit 0x80
            else msg_buf[3] = (msg_buf[3] & 0x80) | 0x10;
        if(addressing_is_word)
        {
          //increment word address only when we have an uneven byte
          if(i & 1)
          {
            address++;
            if((address & 0xFFFF) == 0xFFFF)
              extended_address++;
          }
        } else address++;
        i++;
      }
      //page mode check result:
      // stk sets the Write page bit (7) if the page is complete
      // and we should write it.
      if(msg_buf[3] & 0x80)
      {
        Spi_TxRx(msg_buf[6]);
        Spi_TxRx(HI(saddress));
        Spi_TxRx(LO(saddress));
        Spi_TxRx(0);
        if(!addressing_is_word) Delay_ms(1); // eeprom writing, eeprom needs more time
        //check the different polling mode methods
        ci = 150; // timeout
        if(msg_buf[3] & 0x20 && poll_address)
        {
          //Data value polling
          tmp = msg_buf[8];
          while(tmp == msg_buf[8] && ci)
          {
            // The Low/High byte selection bit is
            // bit number 3. Set high byte for uneven bytes
            // Read data:
            if(poll_address&1)
              Spi_TxRx(msg_buf[7] | (1 << 3));
                else Spi_TxRx(msg_buf[7]);
            Spi_TxRx(HI(poll_address));
            Spi_TxRx(LO(poll_address));
            tmp = Spi_TxRx(0x00);
            ci--;
          }
          if(ci == 0) cstatus = STATUS_CMD_TOUT;
        }
        else if(msg_buf[3] & 0x40)
        {
          //RDY/BSY polling
          do
          {
            ci--;
            Spi_TxRx(0xF0);
            Spi_TxRx(0);
            Spi_TxRx(0);
          }
          while((Spi_TxRx(0) & 1) && ci);
          if(ci == 0) cstatus = STATUS_RDY_BSY_TOUT;
        }
        else
        {
          // simple waiting
          Delay_ms(msg_buf[4]);
        }
      }
    }
    answerlen = 2;
    //msg_buf[0] = CMD_PROGRAM_FLASH_ISP; or CMD_PROGRAM_EEPROM_ISP
    msg_buf[1] = cstatus;
    break;

// ------------------------- CMD_READ_EEPROM_ISP -----------------------------

  case CMD_READ_EEPROM_ISP:
    addressing_is_word = 0;  // address each byte
                
// -------------------------- CMD_READ_FLASH_ISP -----------------------------

  case CMD_READ_FLASH_ISP:
		// msg_buf[1] and msg_buf[2] NumBytes
		// msg_buf[3] cmd
		nbytes = ((unsigned int)msg_buf[1]) << 8;
		nbytes |= msg_buf[2];
		tmp = msg_buf[3];
    if(nbytes > 280) nbytes = 280; // limit answer len, prevent overflow
    i = 0;
    while(i < nbytes)
		{
      Main_Wdt_Reset(1);
      // In commands PROGRAM_FLASH and READ_FLASH "Load Extended Address" 
      // command is executed before every operation if we are programming 
      // processor with Flash memory bigger than 64k words and 64k words boundary 
      // is just crossed or new address was just loaded.
      if(larger_than_64k && ((address & 0xFFFF) == 0 || new_address))
      {
        // load extended addr byte 0x4d
        Spi_TxRx(0x4D);
        Spi_TxRx(0x00);
        Spi_TxRx(extended_address);
        Spi_TxRx(0x00);
        new_address = 0;
      }
			//Select Low or High-Byte
			if(addressing_is_word && i&1)
				Spi_TxRx(tmp | (1 << 3));
			    else Spi_TxRx(tmp);
			Spi_TxRx(HI(address));
			Spi_TxRx(LO(address));
			msg_buf[i + 2] = Spi_TxRx(0);
      if(addressing_is_word)
      {
        //increment word address only when we have an uneven byte 
        if(i & 1)
        {
          address++;
          if((address & 0xFFFF) == 0xFFFF) extended_address++;
        }
      }
      else address++;
      i++;
		}
		answerlen = nbytes+3;
		//msg_buf[0] = CMD_READ_FLASH_ISP; or CMD_READ_EEPROM_ISP
		msg_buf[1] = STATUS_CMD_OK;
		msg_buf[nbytes+2] = STATUS_CMD_OK;
    break;

// ------------------------- CMD_PROGRAM_LOCK_ISP ----------------------------

  case CMD_PROGRAM_LOCK_ISP:

// ------------------------- CMD_PROGRAM_FUSE_ISP ----------------------------
                
  case CMD_PROGRAM_FUSE_ISP:
    Spi_TxRx(msg_buf[1]);
    Spi_TxRx(msg_buf[2]);
    Spi_TxRx(msg_buf[3]);
    Spi_TxRx(msg_buf[4]);
    answerlen =3;
    // msg_buf[0] = CMD_PROGRAM_FUSE_ISP; or CMD_PROGRAM_LOCK_ISP
    msg_buf[1] = STATUS_CMD_OK;
    msg_buf[2] = STATUS_CMD_OK;
    break;

// -------------------------- CMD_READ_OSCCAL_ISP ----------------------------
                
  case CMD_READ_OSCCAL_ISP:

// ------------------------ CMD_READ_SIGNATURE_ISP ---------------------------
                
  case CMD_READ_SIGNATURE_ISP:

// -------------------------- CMD_READ_LOCK_ISP ------------------------------

  case CMD_READ_LOCK_ISP:
                
// -------------------------- CMD_READ_FUSE_ISP ------------------------------

  case CMD_READ_FUSE_ISP:
		for(ci = 0; ci < 4; ci++)
    {
      tmp = Spi_TxRx(msg_buf[ci + 2]);
      if(msg_buf[1] == (ci + 1)) msg_buf[2] = tmp;
			Delay_ms(5);
    }
    answerlen = 4;
    // msg_buf[0] = CMD_READ_FUSE_ISP; or CMD_READ_LOCK_ISP or ...
    msg_buf[1] = STATUS_CMD_OK;
    // msg_buf[2] is the data (fuse byte)
    msg_buf[3] = STATUS_CMD_OK;
    break;

// ---------------------------- CMD_SPI_MULTI --------------------------------

  case CMD_SPI_MULTI:
    // 0: CMD_SPI_MULTI
    // 1: NumTx
    // 2: NumRx 
    // 3: RxStartAddr counting from zero
    // 4+: TxData (len in NumTx)
    // example: 0x1d 0x04 0x04 0x00   0x30 0x00 0x00 0x00
    tmp = msg_buf[2];
    tmp2 = msg_buf[3];
    cj = 0; 
    ci = 0;
    for(cj = 0; cj < msg_buf[1]; cj++)
    {
	    Delay_ms(5);
      if(cj >= tmp2 && ci <tmp)
      {
        // store answer starting from msg_buf[2]
        msg_buf[ci + 2] = Spi_TxRx(msg_buf[cj + 4]);
        ci++;
      }
      else Spi_TxRx(msg_buf[cj + 4]);
    }
    // padd with zero:
    while(ci < tmp)
    {
      msg_buf[ci + 2]=0;
      ci++;
    }
    answerlen = ci + 3;
    // msg_buf[0] = CMD_SPI_MULTI
    msg_buf[1] = STATUS_CMD_OK;
    // msg_buf[2...ci+1] is the data 
    msg_buf[ci+2] = STATUS_CMD_OK;
    break;

    default:
    // we should not come here
    answerlen = 2;
    msg_buf[1] = STATUS_CMD_UNKNOWN;
    break;
  }
  TxAnswer(seqnum, answerlen);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
