// ---------------------------------------------------------------------------

// Project:         STK500 programmer
// Compiler:        IAR EWAVR 5.30
// Microcontroller: ATmega8
// E-mail:          wubblick@yahoo.com

// Previous versions:
// Guido Socher, Copyright: GPL
// Terminal capabilities for updating SW/HW version:
// Idea: Florin-Viorel Petrov
// (www.fonitzu.com/electronics)

// ---------------------------------------------------------------------------

// Modules chain:

// main
//  |--> cmd 
//        |--> terminal
//        |       |------> uart
//        |--------------> uart
//        |---------------------> adc
//        |---------------------------> spi

// ---------------------------------------------------------------------------

#include "main.h"
#include "cmd.h"

// ----------------------------- Constants: ----------------------------------

#define LED_ON_IDLE_TIME    20  // LED on time in IDLE mode, ms
#define LED_OFF_IDLE_TIME 1980  // LED off time in IDLE mode, ms
#define LED_ON_TERM_TIME   500  // LED on time in TERMINAL mode, ms
#define LED_OFF_TERM_TIME  500  // LED on time in TERMINAL mode, ms
#define RESET_TEST_PERIOD  250  // RESET tect period, ms

// ----------------------------- Variables: ----------------------------------

bool tick;                      // system tick

// ------------------------ Function prototypes: -----------------------------

void Main_Wdt_Init(void);       // watchdog timer init
void Main_Ports_Init(void);     // ports init
void Main_Timer_Init(void);     // system timer init
bool Main_Get_Tick(void);
void Main_Led_Control(bool t);  // control the LED

// ---------------------------------------------------------------------------
// ---------------------------- Main program: --------------------------------
// ---------------------------------------------------------------------------

int main(void)
{
  Main_Wdt_Init();              // watchdog init
  Main_Ports_Init();            // ports init
  Main_Timer_Init();            // system timer init
  Cmd_Init();                   // commands parse module init
  __enable_interrupt();         // enable interrupts

  while(1)
  {
    tick = Main_Get_Tick();     // get system tick
    Cmd_Parse(tick);            // parse and execute commands
    Main_Led_Control(tick);     // control the LED
    Main_Wdt_Reset(tick);       // watchdog reset
  }
}

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

// ------------------------ Watchdog timer init: -----------------------------

void Main_Wdt_Init(void)
{
  __watchdog_reset();
  WDTCR = (1 << WDCE) | (1 << WDE);
  WDTCR = (1 << WDE) | (1 << WDP2) | (1 << WDP1) | (1 << WDP0); // time-out 2 s
}

// ------------------------- Watchdog reset: ---------------------------------

void Main_Wdt_Reset(bool t)
{
  if(t)
    __watchdog_reset();
}

// ---------------------------- Ports init: ----------------------------------

void Main_Ports_Init(void)
{
  DDRB  = I_DDRB;
  PORTB = I_PORTB;
  DDRC  = I_DDRC;
  PORTC = I_PORTC;
  DDRD  = I_DDRD;
  PORTD = I_PORTD;
}

// -------------------------- System timer init: -----------------------------

void Main_Timer_Init(void)
{
  // setup timer 2 as system timer:
  TCCR2 = (1 << WGM21) | (1 << CS22) | (1 << CS20); // CTC, CK / 128
  OCR2 = (int)(F_CLK * T_SYS / 128) - 1;
}

// --------------------------- Get system tick: ------------------------------

bool Main_Get_Tick(void)
{
  if(TIFR & (1 << OCF2))
  {
    TIFR = (1 << OCF2);
    return(1);
  }
  return(0);
}

// ---------------------------- LED control: ---------------------------------

void Main_Led_Control(bool t)
{
  static int ResTimer = 0;
  static int LedTimer = 0;
  static bool Active = 0;
  static bool Term = 0;
  static bool LedState = 0;
  
  if(t)
  {
    // poll reset state:
    if(ResTimer) ResTimer--;
    else
    {
      bool a = Cmd_ResetOk();
      bool t = Cmd_TerminalMode();
      if((Active != a) || (Term != t))
      {
        LedTimer = 0;
        LedState = 1;
      }
      Active = a;  
      Term = t;  
      ResTimer = RESET_TEST_PERIOD;
    }
    // LED blinking:
    if(LedTimer) LedTimer--;
    else
    {
      if(LedState)
      {
        LedState = 0;
        if(Cmd_TerminalMode())
          LedTimer = LED_OFF_TERM_TIME;
            else LedTimer = LED_OFF_IDLE_TIME;
      }
      else
      {
        LedState = 1;
        if(Cmd_TerminalMode())
          LedTimer = LED_ON_TERM_TIME;
            else LedTimer = LED_ON_IDLE_TIME;
      }
    }
  }
  // LED on/off:
  if((LedState || (Active && !Term)) && Cmd_ParseState() == MSG_IDLE)
    PORT_LED_1;
      else PORT_LED_0;
}

// -------------------------- Delay (ms range): ------------------------------

void Delay_ms(int d)
{
  while(d)
  {
    Delay_us(1000);     // delay 1 ms
    __watchdog_reset(); // watchdog timer restart
    d--;
  }
}

// ---------------------------------------------------------------------------
