// ---------------------------------------------------------------------------

// STK500 programmer: SPI support module

// ---------------------------------------------------------------------------

#include "main.h"
#include "spi.h"

// ----------------------------- Constants: ----------------------------------

#define DEFAULT_SCK_DURATION 2 // default SCK frequency
// STK-500: 0 - 1.843 MHz, 1 - 460.8 kHz, 2 - 115.2 kHz, 3 - 57.6 kHz
// AVRISP_2: 0 - 921.6 kHz, 1 - 230.4 kHz, 2 - 57.6 kHz,  3 - 28.8 kHz

#define SCK_DIVS   4 // SCK dividers for hardware SPI

//const char __flash SCKDIV[SCK_DIVS] = {4, 16, 64, 128};
const char __flash SCKDIV[SCK_DIVS] = {4, 16, 32, 128}; // hack 115.2 to 230.4

#define SCK_LOOP  26 // min software SCK generation loop duration in cycles
#define SCK_DELAY  6 // additional delay to SCK generation loop in cycles

#define Delay(d) { char __d = d; while(--__d); }

// ----------------------------- Variables: ----------------------------------

static char Sck_Dur;
static char D_Sck_Dur;

// ----------------------------- SPI init: -----------------------------------

void Spi_Init(void)
{
  Spi_SetSckDuration(DEFAULT_SCK_DURATION);
}

// ---------------------------- SPI enable: ----------------------------------

void Spi_Enable(void) 
{
  DIR_RESET_OUT; // reset as output
  PORT_RESET_1;  // reset = high
	DIR_MISO_IN;   // MISO is input
	// now output pins low in case somebody used it as output
	DIR_MOSI_OUT;  // MOSI is output
	PORT_MOSI_0;   // MOSI low
	DIR_SCK_OUT;   // SCK is output
	PORT_SCK_0;    // SCK low
  Delay_ms(10);  // discharge MOSI/SCK
	PORT_RESET_0;  // reset = low, stay active
  Delay_ms(20);  // stab delay
  Spi_RESET_Pulse(); // generate reset pulse
}

// ---------------------------- SPI disable: ---------------------------------

void Spi_Disable(void)
{
	// to be sure:
  DIR_SCK_OUT;   // SCK is output
	PORT_SCK_0;    // SCK low
  PORT_RESET_1;  // reset = high
  // all other pins as input:
	DIR_MISO_IN;   // MISO is input
	PORT_MISO_0;   // pullup off
	DIR_SCK_IN;    // SCK is input
	PORT_SCK_0;    // pullup off
	DIR_MOSI_IN;   // MOSI as input
	PORT_MOSI_0;   // pullup off
  Delay_ms(20); 
  // now we can also set reset high impedence:
	DIR_RESET_IN;  // reset as input
	PORT_RESET_0;  // pullup off
}

//------------------------- Set SCK duration: --------------------------------

void Spi_SetSckDuration(char d)
{
  int div;
  Sck_Dur = d;
  // clock divide ratio for original STK:
  if(d < SCK_DIVS) div = SCKDIV[d];
    else div = (int)d * 24; // according to AVR068
  // clock divide ratio for F_CLK clock frequency: 
  div = (div * (long)(F_CLK * 1E4)) / (long)(F_CLK_STK * 1E4); //1E4 - no OVF
  // limit min divide ratio:
  if(div < SCK_LOOP) div = SCK_LOOP;
  // calculate delay param,
  // SCK duration = SCK_LOOP + SCK_DELAY * (D_Sck_Dur - 1):
  int p = ((div - SCK_LOOP) / SCK_DELAY) + 1;
  // limit max delay param:
  if(p > 255) p = 255;
  D_Sck_Dur = (char)p;
}

//-------------------------- Get SCK duration: -------------------------------

char Spi_GetSckDuration(void)
{
  return(Sck_Dur);
}

// ------------------- Generate pulse at SCK line: ---------------------------

void Spi_SCK_Pulse(void)
{
  PORT_SCK_0;
  Delay(D_Sck_Dur);
  Delay(D_Sck_Dur);
  PORT_SCK_1;
  Delay(D_Sck_Dur);
  Delay(D_Sck_Dur);
  PORT_SCK_0;
}

// ------------------ Generate pulse at RESET line: --------------------------

void Spi_RESET_Pulse(void) 
{
  /* give a positive RESET pulse, we can't guarantee
   * that SCK was low during power up (see Atmel's
   * data sheets, search for "Serial Downloading in e.g atmega8 
   * data sheet):
   * "... the programmer can not guarantee that SCK is held low during
   * Power-up. In this case, RESET must be given a positive pulse of at least two
   * CPU clock cycles duration after SCK has been set to 0." 
   * */
  PORT_RESET_1; // reset = high = not active
  Delay(D_Sck_Dur);
  Delay(D_Sck_Dur);
	PORT_RESET_0; // reset = low, stay active
  Delay_ms(20); // min stab delay
}

//------------------ Send 8 bit, return received byte: -----------------------

char Spi_TxRx(char data)
{
  char i = 0x80;
  char rval = 0;
  char del = D_Sck_Dur; 
    
  if(Sck_Dur == 0)              // if MAX SCK frequency, eliminate delays
  {
    while(i != 0)
    {
      PORT_SCK_0;               // SCK low
      if(data & i) PORT_MOSI_1; // MOSI high
        else PORT_MOSI_0;       // or MOSI low
      if(PIN_MISO) rval |= i;   // read MISO
      PORT_SCK_1;               // SCK high
      __delay_cycles(6);        // make low and high pulse equaly long
      i = i >> 1;
    }
  }
  else
  {
    while(i != 0)
    {
      PORT_SCK_0;               // SCK low
      if(data & i) PORT_MOSI_1; // MOSI high
        else PORT_MOSI_0;       // or MOSI low
      Delay(del);               // delay
      if(PIN_MISO) rval |= i;   // read MISO
      PORT_SCK_1;               // SCK high
      Delay(del);               // delay
      __delay_cycles(5);        // make low and high pulse equaly long
      i = i >> 1;
    }
  }
  PORT_SCK_0;                   // SCK low
  return(rval);
}

/*
// Compiled (optimization: speed, high)

// Sck_Dur = 0:

??Spi_TxRx_0:
    CBI     0x15, 0x04    // 2    SCK = 0
    MOV     R17, R16      // 1
    AND     R17, R19      // 1
    BREQ    ??Spi_TxRx_4  // 2/1
    SBI     0x15, 0x01    // 0/2
    RJMP    ??Spi_TxRx_5  // 0/2
??Spi_TxRx_4:
    CBI     0x15, 0x01    // 2/0
??Spi_TxRx_5:        
    SBIC    0x13, 0x05    // 2/1
    OR      R18, R19      // 0/1
??Spi_TxRx_6:
    SBI     0x15, 0x04    // 2    SCK = 1
    RJMP    $+2           // 2
    RJMP    $+2           // 2
    RJMP    $+2           // 2
    LSR     R19           // 1
    BREQ    ??Spi_TxRx_7  // 1

// Spi_TxRx cycle for minimal delay (D_Sck_Dur = 1):

??Spi_TxRx_0:
    CBI     0x15, 0x04    // 2    SCK = 0
    MOV     R17, R16      // 1
    AND     R17, R19      // 1
    BREQ    ??Spi_TxRx_1  // 2/1
    SBI     0x15, 0x01    // 0/2
    RJMP    ??Spi_TxRx_2  // 0/2
??Spi_TxRx_1:
    CBI     0x15, 0x01    // 2/0
??Spi_TxRx_2:
    MOV     R17, R20      // 1
??Spi_TxRx_3:
    DEC     R17           // 1
    BRNE    ??Spi_TxRx_3  // 1
    SBIC    0x13, 0x05    // 2/1
    OR      R18, R19      // 0/1
    SBI     0x15, 0x04    // 2    SCK = 1
    MOV     R17, R20      // 1
??Spi_TxRx_5:
    DEC     R17           // 1
    BRNE    ??Spi_TxRx_5  // 1
    RJMP    $+2           // 2
    RJMP    $+2           // 2
    NOP                   // 1
    LSR     R19           // 1
    BRNE    ??Spi_TxRx_0  // 2
*/

//----------------------------------------------------------------------------
