// ---------------------------------------------------------------------------

// STK500 programmer: terminal mode support module

// ---------------------------------------------------------------------------

#include "main.h"
#include "terminal.h"
#include "uart.h"

// ----------------------------- Constants: ----------------------------------

#define EEPROM_SIGNATURE           0xBEDA // EEPROM signature
#define D_CONFIG_PARAM_SW_MAJOR         2 // default SW major
#define D_CONFIG_PARAM_SW_MINOR      0x0A // default SW minor
#define DEFAULT_RESET_CHECK             1 // default do reset check

// terminal states:
enum
{
  TS_IDLE,
  TS_WAIT_R,
  TS_GET_SW_MAJOR,
  TS_GET_SW_MINOR,
  TS_GET_RESET_ST
};

// read number states:
enum
{
  RD_DIG1,
  RD_DIG2,
  RD_COMPLETE
};

// ----------------------------- Variables: ----------------------------------

static char TerminalState;      // terminal state
static char ReadNumState;       // read number state
static char ReadNumber;         // read number

static char ConfigParamSWMajor; // current SW major
static char ConfigParamSWMinor; // current SW minor
static char ConfigDoResetCheck; // current do reset check state

__no_init int  __eeprom EepromSignature; // signature in EEPROM
__no_init char __eeprom EepromMajor;     // SW major in EEPROM
__no_init char __eeprom EepromMinor;     // SW minor in EEPROM
__no_init char __eeprom EepromRstCk;     // do reset check state in EEPROM

bool Terminal_ReadNum(char ch); // read number via terminal
char DigL(char d);              // convert LS digit to char
char DigH(char d);              // convert MS digit to char

// --------------------------- Terminal init: --------------------------------

void Terminal_Init(void)
{
  if(EepromSignature == EEPROM_SIGNATURE) // chrck EEPROM signature
  {
    ConfigParamSWMinor = EepromMinor; // read EEPROM if signature OK
    ConfigParamSWMajor = EepromMajor;
    ConfigDoResetCheck = EepromRstCk;
  }
  else
  {
	  ConfigParamSWMinor = D_CONFIG_PARAM_SW_MINOR; // load defaults
	  ConfigParamSWMajor = D_CONFIG_PARAM_SW_MAJOR;
    ConfigDoResetCheck = DEFAULT_RESET_CHECK;
	}
  TerminalState = TS_IDLE;
}

// --------------------------- Terminal mode: --------------------------------

// The special avrusb500 terminal mode.
// Just connect a serial terminal and press enter twice.
// Both Windows and Linux send normally \r as the
// only line end character. It is however possible to configure
// \r\n as line end in some serial terminals. Therefore we ignore \n.

bool Terminal_Exe(char ch)
{
  switch(TerminalState)
  {
  case TS_IDLE:
    if(ch == '\r') TerminalState = TS_WAIT_R; // wait for '\r'
    break;  
  case TS_WAIT_R:
    if(ch == '\n') break; // skip '\n'
    if(ch == '\r') // wait for second '\r'
    {
      Uart_SendStrF("\x1B[0m\x1B[2J\x1B[0;0f"); // init terminal
      Uart_SendStrF("AVRUSB500V3.0");           // send device version
      Uart_SendNewLine();                       // next line
      Uart_SendNewLine();                       // next line
      Uart_SendStrF("Enter SW Version Major in hex [");
      Uart_SendChar(DigH(ConfigParamSWMajor));  // current SW major
      Uart_SendChar(DigL(ConfigParamSWMajor));
      Uart_SendStrF("]: ");
      ReadNumState = RD_DIG1;
      ReadNumber = ConfigParamSWMajor;
      TerminalState = TS_GET_SW_MAJOR;
    }
    else TerminalState = TS_IDLE;
    break;  
  case TS_GET_SW_MAJOR:
    if(Terminal_ReadNum(ch)) // read entered number
    {
      ConfigParamSWMajor = ReadNumber; // new SW major
      EepromMajor = ConfigParamSWMajor;
      Uart_SendNewLine();                       // next line
      Uart_SendStrF("Enter SW Version Minor in hex [");
      Uart_SendChar(DigH(ConfigParamSWMinor));  // current SW minor
      Uart_SendChar(DigL(ConfigParamSWMinor));
      Uart_SendStrF("]: ");
      ReadNumState = RD_DIG1;
      ReadNumber = ConfigParamSWMinor;
      TerminalState = TS_GET_SW_MINOR;
    }
    break;  
  case TS_GET_SW_MINOR:
    if(Terminal_ReadNum(ch)) // read entered number
    {
      ConfigParamSWMinor = ReadNumber; // new SW minor
      EepromMinor = ConfigParamSWMinor;
      Uart_SendNewLine();                       // next line
      Uart_SendStrF("Reset target at connector polarity test (1 = on 0 = off) [");
      Uart_SendChar(DigL(ConfigDoResetCheck));  // current DoReset state
      Uart_SendStrF("]: ");
      ReadNumState = RD_DIG1;
      ReadNumber = ConfigDoResetCheck;
      TerminalState = TS_GET_RESET_ST;
    }
    break;  
  case TS_GET_RESET_ST:
    if(Terminal_ReadNum(ch)) // read entered number
    {
      ConfigDoResetCheck = ReadNumber? 1 : 0; // new DoReset state
      EepromRstCk = ConfigDoResetCheck;
      EepromSignature = EEPROM_SIGNATURE; // data is valid
      Uart_SendNewLine();                       // next line
      Uart_SendNewLine();                       // next line
	    Uart_SendStrF("New SW Version is ");
      Uart_SendChar(DigH(ConfigParamSWMajor));  // new SW major
      Uart_SendChar(DigL(ConfigParamSWMajor));
      Uart_SendChar('.');
      Uart_SendChar(DigH(ConfigParamSWMinor));  // new SW minor
      Uart_SendChar(DigL(ConfigParamSWMinor));
      Uart_SendStrF(" (hex).");
      Uart_SendNewLine();
      Uart_SendStrF("Reset target at connector polarity test is ");
      if(ConfigDoResetCheck == 0)               // new reset test state
        Uart_SendStrF("off.");
          else Uart_SendStrF("on.");
      Uart_SendNewLine();
      Uart_SendStrF("Just close the terminal. No reset needed.");
      Uart_SendNewLine();                       // next line
      TerminalState = TS_IDLE;
    }
    break;
  }
  return(TerminalState > TS_WAIT_R); // return true in terminal mode
}

// ---------- Read a max of 2 hex digits from the serial line ----------------

// return value of 0xff means no change (user just hit return)

bool Terminal_ReadNum(char ch)
{
  char d = 0xFF;
  if((ch >= '0') && (ch <= '9')) d = ch - '0';
  if((ch >= 'A') && (ch <= 'F')) d = ch - 'A' + 10;
  if((ch >= 'a') && (ch <= 'f')) d = ch - 'a' + 10;
  if(d != 0xFF)
  {
    Uart_SendChar(ch); // echo back to terminal
    switch(ReadNumState)
    {
    case RD_DIG1: // most significant digit
      ReadNumber = d;
      ReadNumState = RD_DIG2;
      break;
    case RD_DIG2: // least significant digit
      ReadNumber = (ReadNumber << 4) | d;
      ReadNumState = RD_COMPLETE;
      break;
    }
  }
  if(ch == '\r') return(1);
  return(0);
}
  
// --------------------- Convert LS digit to char:  --------------------------

char DigL(char d)
{
  d = d & 0x0F;
  if(d < 10) return(d + '0');
  return(d + ('A' - 10));
}

// --------------------- Convert MS digit to char:  --------------------------

char DigH(char d)
{
  return(DigL(d >> 4));
}

// ----------------------- Read major SW version -----------------------------

char Terminal_GetSWMajor(void)
{
  return(ConfigParamSWMajor);
}

// ----------------------- Read minor SW version -----------------------------

char Terminal_GetSWMinor(void)
{
  return(ConfigParamSWMinor);
}

// ----------------------- Read Reset Check Mode -----------------------------

char Terminal_GetResetCheck(void)
{
  return(ConfigDoResetCheck);
}

//----------------------------------------------------------------------------
